@@ -229,125 +229,131 @@ struct URITest
   } uri;
 };
 
+#define COMMON_URI_TESTS \
+  /* VALID URIS.  PARSING AND PRINTING OF THESE SHOULD NOT CHANGE */ \
+  /* scheme/path */ \
+  {"scheme:", \
+      {"scheme", NULL, NULL, GST_URI_NO_PORT, NULL, {{NULL, NULL}}, NULL}}, \
+  {"scheme:path", \
+      {"scheme", NULL, NULL, GST_URI_NO_PORT, "path", {{NULL, NULL}}, NULL}}, \
+  {"path", \
+      {NULL, NULL, NULL, GST_URI_NO_PORT, "path", {{NULL, NULL}}, NULL}}, \
+  {"/path", \
+      {NULL, NULL, NULL, GST_URI_NO_PORT, "/path", {{NULL, NULL}}, NULL}}, \
+  /* hostname/port */ \
+  {"scheme://hostname/path", \
+        {"scheme", NULL, "hostname", GST_URI_NO_PORT, "/path", {{NULL, NULL}}, \
+          NULL}}, \
+  {"scheme://hostname:123/path", \
+      {"scheme", NULL, "hostname", 123, "/path", {{NULL, NULL}}, NULL}}, \
+  /* ipv6 hostname/port */ \
+  {"scheme://[01:23:45:67:89:ab:cd:ef]/path", \
+        {"scheme", NULL, "01:23:45:67:89:ab:cd:ef", GST_URI_NO_PORT, "/path", \
+          {{NULL, NULL}}, NULL}}, \
+  {"scheme://[01:23:45:67:89:ab:cd:ef]:123/path", \
+        {"scheme", NULL, "01:23:45:67:89:ab:cd:ef", 123, "/path", {{NULL, \
+                  NULL}}, NULL}}, \
+  /* query/fragment */ \
+  {"path?query", \
+        {NULL, NULL, NULL, GST_URI_NO_PORT, "path", {{"query", NULL}, {NULL, \
+                  NULL}}, NULL}}, \
+  {"path?query=value", \
+        {NULL, NULL, NULL, GST_URI_NO_PORT, "path", {{"query", "value"}, {NULL, \
+                  NULL}}, NULL}}, \
+  {"path?query#fragment", \
+        {NULL, NULL, NULL, GST_URI_NO_PORT, "path", {{"query", NULL}, {NULL, \
+                  NULL}}, "fragment"}}, \
+  {"path?query=value#fragment", \
+        {NULL, NULL, NULL, GST_URI_NO_PORT, "path", {{"query", "value"}, {NULL, \
+                  NULL}}, "fragment"}}, \
+  {"scheme:path?query#fragment", \
+        {"scheme", NULL, NULL, GST_URI_NO_PORT, "path", {{"query", NULL}, {NULL, \
+                  NULL}}, "fragment"}}, \
+  /* full */ \
+  {"scheme://hostname:123/path?query#fragment", \
+        {"scheme", NULL, "hostname", 123, "/path", {{"query", NULL}, {NULL, \
+                  NULL}}, "fragment"}}, \
+  {"scheme://hostname:123/path?query=value#fragment", \
+        {"scheme", NULL, "hostname", 123, "/path", {{"query", "value"}, {NULL, \
+                  NULL}}, "fragment"}}, \
+  {"scheme://hostname:123?query", \
+        {"scheme", NULL, "hostname", 123, NULL, {{"query", NULL}, {NULL, \
+                  NULL}}, NULL}}, \
+  {"scheme://hostname:123?query=value", \
+        {"scheme", NULL, "hostname", 123, NULL, {{"query", "value"}, {NULL, \
+                  NULL}}, NULL}}, \
+  {"scheme://hostname:123?query#fragment", \
+        {"scheme", NULL, "hostname", 123, NULL, {{"query", NULL}, {NULL, \
+                  NULL}}, "fragment"}}, \
+  {"scheme://hostname:123?query=value#fragment", \
+        {"scheme", NULL, "hostname", 123, NULL, {{"query", "value"}, {NULL, \
+                  NULL}}, "fragment"}}, \
+  /* user/pass */ \
+  {"scheme://userinfo@hostname", \
+        {"scheme", "userinfo", "hostname", GST_URI_NO_PORT, NULL, {{NULL, \
+                  NULL}}, NULL}}, \
+  {"scheme://userinfo@hostname:123/path?query#fragment", \
+        {"scheme", "userinfo", "hostname", 123, "/path", {{"query", NULL}, \
+              {NULL, NULL}}, "fragment"}}, \
+  {"scheme://user:pass@hostname", \
+        {"scheme", "user:pass", "hostname", GST_URI_NO_PORT, NULL, {{NULL, \
+                  NULL}}, NULL}}, \
+  {"scheme://user:pass@hostname:123/path?query#fragment", \
+        {"scheme", "user:pass", "hostname", 123, "/path", {{"query", NULL}, \
+              {NULL, NULL}}, "fragment"}}, \
+  /* FUNNY URIS.  PARSING AND PRINTING OF THESE MAY CHANGE */ \
+  {"scheme:hostname:123/path?query#fragment", \
+        {"scheme", NULL, NULL, GST_URI_NO_PORT, "hostname:123/path", {{"query", \
+                  NULL}, {NULL, NULL}}, "fragment"}}, \
+  {"scheme://:pass@hostname:123/path?query#fragment", \
+        {"scheme", ":pass", "hostname", 123, "/path", {{"query", NULL}, {NULL, \
+                  NULL}}, "fragment"}}, \
+  /* Skip initial white space */ \
+  {" \f\n\r\t\vscheme:", \
+      {"scheme", NULL, NULL, GST_URI_NO_PORT, NULL, {{NULL, NULL}}, NULL}}, \
+  {" \f\n\r\t\vpath", \
+      {NULL, NULL, NULL, GST_URI_NO_PORT, "path", {{NULL, NULL}}, NULL}}, \
+  /* file URI */ \
+  {"file://host/home/joe/foo.txt", \
+        {"file", NULL, "host", GST_URI_NO_PORT, "/home/joe/foo.txt", {{NULL, \
+                  NULL}}, NULL}}, \
+  {"file:///home/joe/foo.txt", \
+        {"file", NULL, NULL, GST_URI_NO_PORT, "/home/joe/foo.txt", {{NULL, \
+                  NULL}}, NULL}},
+
+#define UNESCAPED_URI_TESTS \
+  /* Test cases for gst_uri_from_string */ \
+  {"scheme://user%20info@hostname", \
+        {"scheme", "user info", "hostname", GST_URI_NO_PORT, NULL, {{NULL, \
+                  NULL}}, NULL}}, \
+  {"scheme://userinfo@hostname:123/path?query#frag%23ment", \
+        {"scheme", "userinfo", "hostname", 123, "/path", {{"query", NULL}, \
+              {NULL, NULL}}, "frag#ment"}}, \
+  {"scheme://us%3Aer:pass@hostname", \
+        {"scheme", "us:er:pass", "hostname", GST_URI_NO_PORT, NULL, {{NULL, \
+                  NULL}}, NULL}}, \
+  {"scheme://us%3Aer:pa%3Ass@hostname:123/path?query#frag%23ment", \
+        {"scheme", "us:er:pa:ss", "hostname", 123, "/path", {{"query", NULL}, \
+              {NULL, NULL}}, "frag#ment"}},
+
+#define ESCAPED_URI_TESTS \
+  /* Test cases for gst_uri_from_string_escaped */ \
+  {"scheme://user%20info@hostname", \
+        {"scheme", "user%20info", "hostname", GST_URI_NO_PORT, NULL, {{NULL, \
+                  NULL}}, NULL}}, \
+  {"scheme://userinfo@hostname:123/path?query#frag%23ment", \
+        {"scheme", "userinfo", "hostname", 123, "/path", {{"query", NULL}, \
+              {NULL, NULL}}, "frag%23ment"}}, \
+  {"scheme://us%3Aer:pass@hostname", \
+        {"scheme", "us%3Aer:pass", "hostname", GST_URI_NO_PORT, NULL, {{NULL, \
+                  NULL}}, NULL}}, \
+  {"scheme://us%3Aer:pa%3Ass@hostname:123/path?query#frag%23ment", \
+        {"scheme", "us%3Aer:pa%3Ass", "hostname", 123, "/path", {{"query", NULL}, \
+              {NULL, NULL}}, "frag%23ment"}},
+
+
 static const struct URITest tests[] = {
-  /* VALID URIS.  PARSING AND PRINTING OF THESE SHOULD NOT CHANGE */
-
-  /* scheme/path */
-  {"scheme:",
-      {"scheme", NULL, NULL, GST_URI_NO_PORT, NULL, {{NULL, NULL}}, NULL}},
-
-  {"scheme:path",
-      {"scheme", NULL, NULL, GST_URI_NO_PORT, "path", {{NULL, NULL}}, NULL}},
-
-  {"path",
-      {NULL, NULL, NULL, GST_URI_NO_PORT, "path", {{NULL, NULL}}, NULL}},
-
-  {"/path",
-      {NULL, NULL, NULL, GST_URI_NO_PORT, "/path", {{NULL, NULL}}, NULL}},
-
-  /* hostname/port */
-  {"scheme://hostname/path",
-        {"scheme", NULL, "hostname", GST_URI_NO_PORT, "/path", {{NULL, NULL}},
-          NULL}},
-
-  {"scheme://hostname:123/path",
-      {"scheme", NULL, "hostname", 123, "/path", {{NULL, NULL}}, NULL}},
-
-  /* ipv6 hostname/port */
-  {"scheme://[01:23:45:67:89:ab:cd:ef]/path",
-        {"scheme", NULL, "01:23:45:67:89:ab:cd:ef", GST_URI_NO_PORT, "/path",
-          {{NULL, NULL}}, NULL}},
-
-  {"scheme://[01:23:45:67:89:ab:cd:ef]:123/path",
-        {"scheme", NULL, "01:23:45:67:89:ab:cd:ef", 123, "/path", {{NULL,
-                  NULL}}, NULL}},
-
-  /* query/fragment */
-  {"path?query",
-        {NULL, NULL, NULL, GST_URI_NO_PORT, "path", {{"query", NULL}, {NULL,
-                  NULL}}, NULL}},
-  {"path?query=value",
-        {NULL, NULL, NULL, GST_URI_NO_PORT, "path", {{"query", "value"}, {NULL,
-                  NULL}}, NULL}},
-
-  {"path?query#fragment",
-        {NULL, NULL, NULL, GST_URI_NO_PORT, "path", {{"query", NULL}, {NULL,
-                  NULL}}, "fragment"}},
-
-  {"path?query=value#fragment",
-        {NULL, NULL, NULL, GST_URI_NO_PORT, "path", {{"query", "value"}, {NULL,
-                  NULL}}, "fragment"}},
-
-  {"scheme:path?query#fragment",
-        {"scheme", NULL, NULL, GST_URI_NO_PORT, "path", {{"query", NULL}, {NULL,
-                  NULL}}, "fragment"}},
-
-  /* full */
-  {"scheme://hostname:123/path?query#fragment",
-        {"scheme", NULL, "hostname", 123, "/path", {{"query", NULL}, {NULL,
-                  NULL}}, "fragment"}},
-
-  {"scheme://hostname:123/path?query=value#fragment",
-        {"scheme", NULL, "hostname", 123, "/path", {{"query", "value"}, {NULL,
-                  NULL}}, "fragment"}},
-
-  {"scheme://hostname:123?query",
-        {"scheme", NULL, "hostname", 123, NULL, {{"query", NULL}, {NULL,
-                  NULL}}, NULL}},
-
-  {"scheme://hostname:123?query=value",
-        {"scheme", NULL, "hostname", 123, NULL, {{"query", "value"}, {NULL,
-                  NULL}}, NULL}},
-
-  {"scheme://hostname:123?query#fragment",
-        {"scheme", NULL, "hostname", 123, NULL, {{"query", NULL}, {NULL,
-                  NULL}}, "fragment"}},
-
-  {"scheme://hostname:123?query=value#fragment",
-        {"scheme", NULL, "hostname", 123, NULL, {{"query", "value"}, {NULL,
-                  NULL}}, "fragment"}},
-
-  /* user/pass */
-  {"scheme://userinfo@hostname",
-        {"scheme", "userinfo", "hostname", GST_URI_NO_PORT, NULL, {{NULL,
-                  NULL}}, NULL}},
-
-  {"scheme://userinfo@hostname:123/path?query#fragment",
-        {"scheme", "userinfo", "hostname", 123, "/path", {{"query", NULL},
-              {NULL, NULL}}, "fragment"}},
-
-  {"scheme://user:pass@hostname",
-        {"scheme", "user:pass", "hostname", GST_URI_NO_PORT, NULL, {{NULL,
-                  NULL}}, NULL}},
-
-  {"scheme://user:pass@hostname:123/path?query#fragment",
-        {"scheme", "user:pass", "hostname", 123, "/path", {{"query", NULL},
-              {NULL, NULL}}, "fragment"}},
-
-  /* FUNNY URIS.  PARSING AND PRINTING OF THESE MAY CHANGE */
-
-  {"scheme:hostname:123/path?query#fragment",
-        {"scheme", NULL, NULL, GST_URI_NO_PORT, "hostname:123/path", {{"query",
-                  NULL}, {NULL, NULL}}, "fragment"}},
-
-  {"scheme://:pass@hostname:123/path?query#fragment",
-        {"scheme", ":pass", "hostname", 123, "/path", {{"query", NULL}, {NULL,
-                  NULL}}, "fragment"}},
-
-  /* Skip initial white space */
-  {" \f\n\r\t\vscheme:",
-      {"scheme", NULL, NULL, GST_URI_NO_PORT, NULL, {{NULL, NULL}}, NULL}},
-
-  {" \f\n\r\t\vpath",
-      {NULL, NULL, NULL, GST_URI_NO_PORT, "path", {{NULL, NULL}}, NULL}},
-
-  /* file URI */
-  {"file://host/home/joe/foo.txt",
-        {"file", NULL, "host", GST_URI_NO_PORT, "/home/joe/foo.txt", {{NULL,
-                  NULL}}, NULL}},
-  {"file:///home/joe/foo.txt",
-        {"file", NULL, NULL, GST_URI_NO_PORT, "/home/joe/foo.txt", {{NULL,
-                  NULL}}, NULL}},
+  COMMON_URI_TESTS UNESCAPED_URI_TESTS
 };
 
 static const gchar *unparsable_uri_tests[] = {
@@ -409,17 +415,69 @@ GST_START_TEST (test_url_parsing)
 
 GST_END_TEST;
 
+
+static const struct URITest escaped_tests[] = {
+  COMMON_URI_TESTS ESCAPED_URI_TESTS
+};
+
+GST_START_TEST (test_url_parsing_escaped)
+{
+  GstUri *uri;
+  GList *list;
+  gchar *tmp_str;
+  guint i, j;
+
+  for (i = 0; i < G_N_ELEMENTS (escaped_tests); i++) {
+    GST_DEBUG ("Testing URI '%s'", escaped_tests[i].str);
+
+    uri = gst_uri_from_string_escaped (escaped_tests[i].str);
+    fail_unless (uri != NULL);
+    fail_unless_equals_string (gst_uri_get_scheme (uri),
+        escaped_tests[i].uri.scheme);
+    fail_unless_equals_string (gst_uri_get_userinfo (uri),
+        escaped_tests[i].uri.userinfo);
+    fail_unless_equals_string (gst_uri_get_host (uri),
+        escaped_tests[i].uri.host);
+    fail_unless_equals_int (gst_uri_get_port (uri), escaped_tests[i].uri.port);
+    tmp_str = gst_uri_get_path (uri);
+    fail_unless_equals_string (tmp_str, escaped_tests[i].uri.path);
+    g_free (tmp_str);
+
+    for (j = 0; j < 10; j++) {
+      if (!escaped_tests[i].uri.query[j].key)
+        break;
+
+      if (escaped_tests[i].uri.query[j].value) {
+        fail_unless_equals_string (gst_uri_get_query_value (uri,
+                escaped_tests[i].uri.query[j].key),
+            escaped_tests[i].uri.query[j].value);
+      } else {
+        fail_unless (gst_uri_query_has_key (uri,
+                escaped_tests[i].uri.query[j].key));
+      }
+    }
+    list = gst_uri_get_query_keys (uri);
+    fail_unless_equals_int (j, g_list_length (list));
+    g_list_free (list);
+    gst_uri_unref (uri);
+  }
+
+  for (i = 0; i < G_N_ELEMENTS (unparsable_uri_tests); i++) {
+    GST_DEBUG ("Testing unparsable URI '%s'", unparsable_uri_tests[i]);
+
+    uri = gst_uri_from_string (unparsable_uri_tests[i]);
+    fail_unless (uri == NULL);
+  }
+}
+
+GST_END_TEST;
+
 static const struct URITest url_presenting_tests[] = {
   /* check all URI elements present */
   {.uri = {"scheme", "user:pass", "host", 1234, "/path/to/dir",
           {{"query", NULL}, {"key", "value"}}, "fragment"},
       .str =
-#if GLIB_CHECK_VERSION(2, 59, 0)
       "scheme://user:pass@host:1234/path/to/dir?key=value&query#fragment"},
-#else
-      "scheme://user:pass@host:1234/path/to/dir?query&key=value#fragment"},
-#endif
-
   /* IPv6 literal should render in square brackets */
   {.uri = {"scheme", "user:pass", "12:34:56:78:9a:bc:de:f0", 1234,
           "/path/to/dir", {{"query", "value"}}, "fragment"},
@@ -981,24 +1039,63 @@ GST_START_TEST (test_url_get_set)
 
   fail_unless (gst_uri_set_query_value (url, "key", "value"));
   tmp_str = gst_uri_to_string (url);
-#if GLIB_CHECK_VERSION(2, 59, 0)
   fail_unless_equals_string (tmp_str,
       "//example.com/path/to/file/there/segment?key=value&query#fragment");
-#else
+  g_free (tmp_str);
+
+  /* query string (full) */
+  tmp_str = gst_uri_get_query_string (url);
+  fail_unless_equals_string (tmp_str, "key=value&query");
+  g_free (tmp_str);
+
+  /* query string (single key, no value) */
+  tmp_list = g_list_append (NULL, g_strdup ("query"));
+  tmp_str = gst_uri_get_query_string_ordered (url, tmp_list);
+  fail_unless_equals_string (tmp_str, "query");
+  g_free (tmp_str);
+  tmp_str = gst_uri_to_string_with_keys (url, tmp_list);
+  fail_unless_equals_string (tmp_str,
+      "//example.com/path/to/file/there/segment?query#fragment");
+  g_free (tmp_str);
+  g_list_free_full (tmp_list, g_free);
+
+  /* query string (single key, with value) */
+  tmp_list = g_list_append (NULL, g_strdup ("key"));
+  tmp_str = gst_uri_get_query_string_ordered (url, tmp_list);
+  fail_unless_equals_string (tmp_str, "key=value");
+  g_free (tmp_str);
+  tmp_str = gst_uri_to_string_with_keys (url, tmp_list);
+  fail_unless_equals_string (tmp_str,
+      "//example.com/path/to/file/there/segment?key=value#fragment");
+  g_free (tmp_str);
+  g_list_free_full (tmp_list, g_free);
+
+  /* query string (key not present) */
+  tmp_list = g_list_append (NULL, g_strdup ("absent"));
+  tmp_str = gst_uri_get_query_string_ordered (url, tmp_list);
+  fail_if (tmp_str != NULL);
+  tmp_str = gst_uri_to_string_with_keys (url, tmp_list);
+  fail_unless_equals_string (tmp_str,
+      "//example.com/path/to/file/there/segment#fragment");
+  g_free (tmp_str);
+  g_list_free_full (tmp_list, g_free);
+
+  /* query string (both keys, inverse order) */
+  tmp_list = g_list_append (NULL, g_strdup ("query"));
+  tmp_list = g_list_append (tmp_list, g_strdup ("key"));
+  tmp_str = gst_uri_get_query_string_ordered (url, tmp_list);
+  fail_unless_equals_string (tmp_str, "query&key=value");
+  g_free (tmp_str);
+  tmp_str = gst_uri_to_string_with_keys (url, tmp_list);
   fail_unless_equals_string (tmp_str,
       "//example.com/path/to/file/there/segment?query&key=value#fragment");
-#endif
   g_free (tmp_str);
+  g_list_free_full (tmp_list, g_free);
 
   fail_unless (gst_uri_set_query_value (url, "key", NULL));
   tmp_str = gst_uri_to_string (url);
-#if GLIB_CHECK_VERSION(2, 59, 0)
   fail_unless_equals_string (tmp_str,
       "//example.com/path/to/file/there/segment?key&query#fragment");
-#else
-  fail_unless_equals_string (tmp_str,
-      "//example.com/path/to/file/there/segment?query&key#fragment");
-#endif
   g_free (tmp_str);
 
   fail_unless (!gst_uri_set_query_value (NULL, "key", "value"));
@@ -1108,6 +1205,28 @@ GST_START_TEST (test_url_get_media_fragment_table)
 
 GST_END_TEST;
 
+GST_START_TEST (test_url_unescape_equals_in_http_query)
+{
+  GstUri *url;
+  gchar *query_string;
+
+  url =
+      gst_uri_from_string
+      ("http://abc.manifest?token=exp=123~acl=/QualityLevels(*~hmac=0cb");
+
+  fail_unless_equals_string (gst_uri_get_scheme (url), "http");
+  query_string = gst_uri_get_query_string (url);
+  fail_unless_equals_string (query_string,
+      "token=exp=123~acl=/QualityLevels(*~hmac=0cb");
+  g_free (query_string);
+  fail_unless (gst_uri_query_has_key (url, "token"));
+  fail_unless_equals_string (gst_uri_get_query_value (url, "token"),
+      "exp=123~acl=/QualityLevels(*~hmac=0cb");
+  gst_uri_unref (url);
+}
+
+GST_END_TEST;
+
 static Suite *
 gst_uri_suite (void)
 {
@@ -1128,6 +1247,7 @@ gst_uri_suite (void)
   tcase_add_test (tc_chain, test_win32_uri);
 #endif
   tcase_add_test (tc_chain, test_url_parsing);
+  tcase_add_test (tc_chain, test_url_parsing_escaped);
   tcase_add_test (tc_chain, test_url_presenting);
   tcase_add_test (tc_chain, test_url_normalization);
   tcase_add_test (tc_chain, test_url_joining);
@@ -1135,6 +1255,7 @@ gst_uri_suite (void)
   tcase_add_test (tc_chain, test_url_constructors);
   tcase_add_test (tc_chain, test_url_get_set);
   tcase_add_test (tc_chain, test_url_get_media_fragment_table);
+  tcase_add_test (tc_chain, test_url_unescape_equals_in_http_query);
 
   return s;
 }
